#include "resources.h"
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>

#if 1
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
void myLog(const char *fmt, ...){
    va_list args;
    va_start(args, fmt);
    char * newfmt;
    struct timespec tp;
    clock_gettime(CLOCK_REALTIME, &tp);
    asprintf(&newfmt, "%ld.%09ld[%lx] %s\n", tp.tv_sec, tp.tv_nsec, pthread_self(), fmt);
    vfprintf(stderr, newfmt, args);
    free(newfmt);
}
#else
void myLog(const char *fmt, ...) {}
#endif

struct who {
    pthread_t tid;
    int cnt;
    struct who *next;
};

pthread_mutex_t resMtx = PTHREAD_MUTEX_INITIALIZER;
int resMaxIdx;
struct res {
    pthread_mutex_t mtx;
    pthread_cond_t cv;
    int avail;
    int total;
    struct who *wants;
    struct who *has;
} **res;

struct who *append(struct who **b) {
    if (!*b) {
        *b = malloc(sizeof(struct who));
        (*b)->next = NULL;
        return *b;
    }
    struct who *it = *b;
    while (it->next)
        it++;
    it->next = malloc(sizeof(struct who));
    it->next->next = NULL;
    return it->next;
}

void resources_init() {
    myLog("init");
    resMaxIdx = 16;
    res = malloc((resMaxIdx + 1) * sizeof(struct res *));
    for (int i = 0; i <= resMaxIdx; ++i)
        res[i] = NULL;
}

int resources_create(int identifier, int availableCount) {
    myLog("create %d %d", identifier, availableCount);
    if (identifier < 0 || availableCount <= 0)
        return 1;
    pthread_mutex_lock(&resMtx);

    // grow if needed
    if (identifier > resMaxIdx) {
        res = realloc(res, (identifier + 1) * sizeof(struct res *));
        for (int i = resMaxIdx + 1; i <= identifier; ++i)
            res[i] = NULL;
        resMaxIdx = identifier;
    }

    if (res[identifier]) {
        return 2;
    }

    res[identifier] = malloc(sizeof(struct res));

    pthread_mutex_init(&res[identifier]->mtx, NULL);
    pthread_cond_init(&res[identifier]->cv, NULL);
    res[identifier]->total = availableCount;
    res[identifier]->avail = availableCount;
    res[identifier]->wants = NULL;
    res[identifier]->has = NULL;

    pthread_mutex_unlock(&resMtx);
    return 0;
}

int cmpRR(const void *a, const void *b) {
    return ((const struct resources_request *)a)->kind - ((const struct resources_request *)b)->kind;
}

int resources_take(int arrayLength, struct resources_request *what) {
    myLog("take %d elems", arrayLength);
    if (arrayLength < 0)
        return 1;
    if (arrayLength == 0)
        return 0;

    qsort(what, arrayLength, sizeof(struct resources_request), cmpRR);

    for (int i = 1; i < arrayLength; ++i)
        if (what[i - 1].kind == what[i].kind)
            return 2;

    struct res *wanted[arrayLength];

    pthread_mutex_lock(&resMtx);
    // verify if thread has no res by looking through all res (inefficient)
    for (int i = 0; i <= resMaxIdx; ++i) {
        if (!res[i])
            continue;
        struct who *it = res[i]->has;
        while (it) {
            if (it->tid == pthread_self()) {
                pthread_mutex_unlock(&resMtx);
                return 3;
            }
            it = it->next;
        }
    }
    // collect wanted
    for (int i = 0; i < arrayLength; ++i) {
        myLog("take wanting id %d cnt %d ", what[i].kind, what[i].count);
        if (!res[what[i].kind]) {
            pthread_mutex_unlock(&resMtx);
            return 4;
        }
        if (res[what[i].kind]->total < what[i].count) {
            pthread_mutex_unlock(&resMtx);
            return 5;
        }
        wanted[i] = res[what[i].kind];
    }
    pthread_mutex_unlock(&resMtx);

    // mark intent
    for (int i = 0; i < arrayLength; ++i)
        pthread_mutex_lock(&wanted[i]->mtx);
    for (int i = 0; i < arrayLength; ++i) {
        myLog("take marking %d", i);
        struct who *me = append(&wanted[i]->wants);
        me->tid = pthread_self();
        me->cnt = what[i].count;
    }
    for (int i = 0; i < arrayLength; ++i)
        pthread_mutex_unlock(&wanted[i]->mtx);

    // grab resources
    for (int i = 0; i < arrayLength; ++i) {
        myLog("take grabbing %d", i);
        pthread_mutex_lock(&wanted[i]->mtx);
        struct who *it, *prev;
        while (1) {
            // count all resources waited for in queue up to including me
            int req = 0;
            prev = NULL;
            it = wanted[i]->wants;
            do {
                if (prev) {
                    prev = it;
                    it = it->next;
                } else
                    prev = it;
                req += it->cnt;
            } while (it->tid != pthread_self());
            // if there are at least as many avail, the proceed
            if (req <= wanted[i]->avail)
                break;
            myLog("take condwaiting %d", i);
            pthread_cond_wait(&wanted[i]->cv, &wanted[i]->mtx);
        }
        myLog("take grabbed %d", i);
        wanted[i]->avail -= what[i].count;
        // move self, in the resource, from wants to has
        if (it == wanted[i]->wants)
            wanted[i]->wants = it->next;
        else
            prev->next = it->next;
        it->next = wanted[i]->has;
        wanted[i]->has = it;
        pthread_mutex_unlock(&wanted[i]->mtx);
    }
    return 0;
}

int resources_return(int identifier, int count) {
    myLog("return %d cnt %d ", identifier, count);
    pthread_mutex_lock(&resMtx);
    struct res *r = res[identifier];
    pthread_mutex_unlock(&resMtx);
    if (!r)
        return 1;

    pthread_mutex_lock(&r->mtx);

    struct who *prev = NULL, *it = r->has;
    while (1) {
        if (prev) {
            prev = it;
            it = it->next;
        } else
            prev = it;

        if (!it) {
            pthread_mutex_unlock(&r->mtx);
            return 2;
        }
        if (it->tid == pthread_self()) {
            if (it->cnt < count) {
                pthread_mutex_unlock(&r->mtx);
                return 3;
            }
            it->cnt -= count;
            r->avail += count;
            if (it->cnt == 0) {
                if (r->has == it)
                    r->has = it->next;
                else
                    prev->next = it->next;
                free(it);
            }
            pthread_cond_broadcast(&r->cv);
            pthread_mutex_unlock(&r->mtx);
            return 0;
        }
    }
}
